"""Tests for messages.py - XBus message handling."""

import pytest
from z21aio.messages import (
    XBusMessage,
    LAN_GET_SERIAL_NUMBER,
    LAN_XBUS_HEADER,
    XBUS_SET_TRACK_POWER,
    XBUS_LOCO_DRIVE,
    XBUS_LOCO_GET_INFO,
    XBUS_GET_VERSION,
    XBUS_GET_FIRMWARE_VERSION,
    XBUS_TURNOUT_INFO,
    XBUS_SET_TURNOUT,
)
from z21aio.types import DccThrottleSteps, FunctionAction


class TestXBusMessage:
    """Tests for XBusMessage class."""

    def test_xor_calculation_simple(self):
        """Test XOR checksum calculation with simple data."""
        msg = XBusMessage(x_header=0x21, dbs=b"\x81")
        # XOR of 0x21 and 0x81 = 0xA0
        assert msg.xor == 0xA0

    def test_xor_calculation_multiple_bytes(self):
        """Test XOR checksum with multiple data bytes."""
        msg = XBusMessage(x_header=0xE3, dbs=b"\xF0\x00\x03")
        # 0xE3 ^ 0xF0 ^ 0x00 ^ 0x03 = 0x10
        assert msg.xor == 0x10

    def test_xor_calculation_empty_data(self):
        """Test XOR checksum with no data bytes."""
        msg = XBusMessage(x_header=0x21, dbs=b"")
        assert msg.xor == 0x21

    def test_to_bytes(self):
        """Test serialization to bytes."""
        msg = XBusMessage(x_header=0x21, dbs=b"\x81")
        result = msg.to_bytes()

        assert result == b"\x21\x81\xa0"

    def test_from_bytes_valid(self):
        """Test parsing from valid bytes."""
        data = b"\x21\x81\xa0"
        msg = XBusMessage.from_bytes(data)

        assert msg.x_header == 0x21
        assert msg.dbs == b"\x81"
        assert msg.xor == 0xA0

    def test_from_bytes_invalid_xor(self):
        """Test that invalid XOR raises ValueError."""
        data = b"\x21\x81\xFF"  # Wrong XOR
        with pytest.raises(ValueError, match="XOR mismatch"):
            XBusMessage.from_bytes(data)

    def test_from_bytes_too_short(self):
        """Test that too-short data raises ValueError."""
        with pytest.raises(ValueError, match="at least 2 bytes"):
            XBusMessage.from_bytes(b"\x21")

    def test_roundtrip(self):
        """Test serialization and deserialization roundtrip."""
        original = XBusMessage(x_header=0xE4, dbs=b"\x13\x00\x03\x64")
        serialized = original.to_bytes()
        restored = XBusMessage.from_bytes(serialized)

        assert restored.x_header == original.x_header
        assert restored.dbs == original.dbs

    def test_track_power_on(self):
        """Test track power on command creation."""
        msg = XBusMessage.track_power_on()

        assert msg.x_header == XBUS_SET_TRACK_POWER
        assert msg.dbs == b"\x81"

    def test_track_power_off(self):
        """Test track power off command creation."""
        msg = XBusMessage.track_power_off()

        assert msg.x_header == XBUS_SET_TRACK_POWER
        assert msg.dbs == b"\x80"

    def test_get_firmware_version(self):
        """Test get firmware version command creation."""
        msg = XBusMessage.get_firmware_version()

        assert msg.x_header == XBUS_GET_FIRMWARE_VERSION
        assert msg.dbs == b"\x0a"
        # XOR = 0xF1 ^ 0x0A = 0xFB
        assert msg.xor == 0xFB

    def test_get_version(self):
        """Test get X-BUS version command creation."""
        msg = XBusMessage.get_version()

        assert msg.x_header == XBUS_GET_VERSION
        assert msg.dbs == b"\x21"
        # XOR = 0x21 ^ 0x21 = 0x00
        assert msg.xor == 0x00

    def test_loco_get_info_short_address(self):
        """Test loco get info command for address < 128."""
        msg = XBusMessage.loco_get_info(3)

        assert msg.x_header == XBUS_LOCO_GET_INFO
        assert msg.dbs[0] == 0xF0
        assert msg.dbs[1] == 0x00  # MSB
        assert msg.dbs[2] == 0x03  # LSB

    def test_loco_get_info_long_address(self):
        """Test loco get info command for address >= 128."""
        msg = XBusMessage.loco_get_info(1234)

        assert msg.x_header == XBUS_LOCO_GET_INFO
        assert msg.dbs[0] == 0xF0
        # Address 1234 = 0x04D2
        # With 0xC0 flag: MSB = 0xC4, LSB = 0xD2
        assert msg.dbs[1] == 0xC4
        assert msg.dbs[2] == 0xD2

    def test_loco_drive_128_steps_forward(self):
        """Test drive command with 128 steps, forward."""
        msg = XBusMessage.loco_drive(3, DccThrottleSteps.STEPS_128, 0xC0)  # 64 forward

        assert msg.x_header == XBUS_LOCO_DRIVE
        assert msg.dbs[0] == 0x13  # 128 steps
        assert msg.dbs[1] == 0x00  # Address MSB
        assert msg.dbs[2] == 0x03  # Address LSB
        assert msg.dbs[3] == 0xC0  # Speed byte

    def test_loco_drive_28_steps_reverse(self):
        """Test drive command with 28 steps, reverse."""
        msg = XBusMessage.loco_drive(5, DccThrottleSteps.STEPS_28, 0x14)  # 20 reverse

        assert msg.x_header == XBUS_LOCO_DRIVE
        assert msg.dbs[0] == 0x12  # 28 steps
        assert msg.dbs[3] == 0x14

    def test_loco_drive_14_steps(self):
        """Test drive command with 14 steps."""
        msg = XBusMessage.loco_drive(10, DccThrottleSteps.STEPS_14, 0x87)  # 7 forward

        assert msg.x_header == XBUS_LOCO_DRIVE
        assert msg.dbs[0] == 0x10  # 14 steps

    def test_loco_drive_long_address(self):
        """Test drive command with long address."""
        msg = XBusMessage.loco_drive(500, DccThrottleSteps.STEPS_128, 0x80)

        # Address 500 = 0x01F4
        # With 0xC0 flag: MSB = 0xC1, LSB = 0xF4
        assert msg.dbs[1] == 0xC1
        assert msg.dbs[2] == 0xF4

    def test_loco_function_on(self):
        """Test function on command."""
        msg = XBusMessage.loco_function(3, 0, FunctionAction.ON)

        assert msg.x_header == XBUS_LOCO_DRIVE
        assert msg.dbs[0] == 0xF8
        # Function byte: 01 000000 = 0x40
        assert msg.dbs[3] == 0x40

    def test_loco_function_off(self):
        """Test function off command."""
        msg = XBusMessage.loco_function(3, 5, FunctionAction.OFF)

        assert msg.dbs[0] == 0xF8
        # Function byte: 00 000101 = 0x05
        assert msg.dbs[3] == 0x05

    def test_loco_function_toggle(self):
        """Test function toggle command."""
        msg = XBusMessage.loco_function(3, 10, FunctionAction.TOGGLE)

        # Function byte: 10 001010 = 0x8A
        assert msg.dbs[3] == 0x8A

    def test_loco_function_invalid_index(self):
        """Test that invalid function index raises ValueError."""
        with pytest.raises(ValueError, match="0-31"):
            XBusMessage.loco_function(3, 32, FunctionAction.ON)

        with pytest.raises(ValueError, match="0-31"):
            XBusMessage.loco_function(3, -1, FunctionAction.ON)

    def test_loco_function_all_functions(self):
        """Test all function indices 0-31."""
        for i in range(32):
            msg = XBusMessage.loco_function(1, i, FunctionAction.ON)
            # Function byte should have function number in lower 6 bits
            assert (msg.dbs[3] & 0x3F) == i
            # Action ON = 01 in upper 2 bits
            assert (msg.dbs[3] & 0xC0) == 0x40

    def test_get_turnout_info(self):
        """Test get turnout info command for address 0."""
        msg = XBusMessage.get_turnout_info(0)

        assert msg.x_header == XBUS_TURNOUT_INFO
        assert msg.dbs[0] == 0x00  # MSB
        assert msg.dbs[1] == 0x00  # LSB

    def test_get_turnout_info_high_address(self):
        """Test get turnout info command for address > 255."""
        msg = XBusMessage.get_turnout_info(500)

        assert msg.x_header == XBUS_TURNOUT_INFO
        # Address 500 = 0x01F4
        assert msg.dbs[0] == 0x01  # MSB
        assert msg.dbs[1] == 0xF4  # LSB

    def test_set_turnout_activate_p0_queue(self):
        """Test set turnout: activate output 0, queue mode."""
        msg = XBusMessage.set_turnout(0, output=0, activate=True, queue_mode=True)

        assert msg.x_header == XBUS_SET_TURNOUT
        # DB2: 1_1_0_1_0_0_0 = 0xA8
        assert msg.dbs[2] == 0xA8

    def test_set_turnout_activate_p1_queue(self):
        """Test set turnout: activate output 1, queue mode."""
        msg = XBusMessage.set_turnout(0, output=1, activate=True, queue_mode=True)

        assert msg.x_header == XBUS_SET_TURNOUT
        # DB2: 1_1_0_1_0_0_1 = 0xA9
        assert msg.dbs[2] == 0xA9

    def test_set_turnout_deactivate_p0_queue(self):
        """Test set turnout: deactivate output 0, queue mode."""
        msg = XBusMessage.set_turnout(0, output=0, activate=False, queue_mode=True)

        assert msg.x_header == XBUS_SET_TURNOUT
        # DB2: 1_1_0_0_0_0_0 = 0xA0
        assert msg.dbs[2] == 0xA0

    def test_set_turnout_activate_p0_immediate(self):
        """Test set turnout: activate output 0, immediate mode."""
        msg = XBusMessage.set_turnout(0, output=0, activate=True, queue_mode=False)

        assert msg.x_header == XBUS_SET_TURNOUT
        # DB2: 1_0_0_1_0_0_0 = 0x88
        assert msg.dbs[2] == 0x88

    def test_set_turnout_deactivate_p1_immediate(self):
        """Test set turnout: deactivate output 1, immediate mode."""
        msg = XBusMessage.set_turnout(0, output=1, activate=False, queue_mode=False)

        assert msg.x_header == XBUS_SET_TURNOUT
        # DB2: 1_0_0_0_0_0_1 = 0x81
        assert msg.dbs[2] == 0x81

    def test_set_turnout_address_encoding(self):
        """Test set turnout address encoding."""
        msg = XBusMessage.set_turnout(500, output=0, activate=True)

        # Address 500 = 0x01F4
        assert msg.dbs[0] == 0x01  # MSB
        assert msg.dbs[1] == 0xF4  # LSB

    def test_set_turnout_invalid_output(self):
        """Test that invalid output raises ValueError."""
        with pytest.raises(ValueError, match="0 or 1"):
            XBusMessage.set_turnout(0, output=2, activate=True)

    def test_repr(self):
        """Test string representation."""
        msg = XBusMessage(x_header=0x21, dbs=b"\x81")
        repr_str = repr(msg)

        assert "0x21" in repr_str.lower()
        assert "81" in repr_str.lower()
        assert "0xa0" in repr_str.lower()



"""Tests for turnout.py - Turnout control and TurnoutState/TurnoutPosition types."""

import pytest
from unittest.mock import AsyncMock, MagicMock, patch

from z21aio.messages import XBUS_TURNOUT_INFO, XBUS_SET_TURNOUT, XBusMessage
from z21aio.packet import Packet
from z21aio.turnout import Turnout
from z21aio.types import TurnoutPosition, TurnoutState


class TestTurnoutPosition:
    """Tests for TurnoutPosition enum."""

    def test_enum_values(self):
        """Test enum integer values match protocol ZZ bits."""
        assert TurnoutPosition.UNKNOWN == 0
        assert TurnoutPosition.P0 == 1
        assert TurnoutPosition.P1 == 2
        assert TurnoutPosition.INVALID == 3


class TestTurnoutState:
    """Tests for TurnoutState dataclass."""

    def test_from_bytes_position_p0(self):
        """Test parsing ZZ=01 as P0."""
        state = TurnoutState.from_bytes(bytes([0x00, 0x00, 0x01]))

        assert state.address == 0
        assert state.position == TurnoutPosition.P0

    def test_from_bytes_position_p1(self):
        """Test parsing ZZ=10 as P1."""
        state = TurnoutState.from_bytes(bytes([0x00, 0x00, 0x02]))

        assert state.address == 0
        assert state.position == TurnoutPosition.P1

    def test_from_bytes_unknown(self):
        """Test parsing ZZ=00 as UNKNOWN."""
        state = TurnoutState.from_bytes(bytes([0x00, 0x00, 0x00]))

        assert state.position == TurnoutPosition.UNKNOWN

    def test_from_bytes_invalid(self):
        """Test parsing ZZ=11 as INVALID."""
        state = TurnoutState.from_bytes(bytes([0x00, 0x00, 0x03]))

        assert state.position == TurnoutPosition.INVALID

    def test_from_bytes_address_encoding(self):
        """Test address is (MSB << 8) | LSB."""
        state = TurnoutState.from_bytes(bytes([0x01, 0xF4, 0x01]))

        assert state.address == 500

    def test_from_bytes_masks_zz_bits(self):
        """Test that only lower 2 bits of status byte are used."""
        state = TurnoutState.from_bytes(bytes([0x00, 0x00, 0xFE]))

        # 0xFE & 0x03 = 0x02 = P1
        assert state.position == TurnoutPosition.P1

    def test_from_bytes_wrong_length(self):
        """Test that wrong length raises ValueError."""
        with pytest.raises(ValueError, match="3 bytes"):
            TurnoutState.from_bytes(bytes([0x00, 0x00]))

        with pytest.raises(ValueError, match="3 bytes"):
            TurnoutState.from_bytes(bytes([0x00, 0x00, 0x00, 0x00]))


class TestTurnout:
    """Tests for Turnout class."""

    @pytest.fixture
    def mock_station(self):
        """Create a mock Z21Station."""
        station = MagicMock()
        station.send_xbus_command = AsyncMock()
        station._subscribers = {}
        return station

    def test_init(self, mock_station):
        """Test Turnout initialization."""
        turnout = Turnout(mock_station, address=5)

        assert turnout.address == 5

    @pytest.mark.asyncio
    async def test_control_returns_turnout(self, mock_station):
        """Test that control() returns a Turnout with the correct address."""
        response_msg = XBusMessage(
            x_header=XBUS_TURNOUT_INFO,
            dbs=bytes([0x00, 0x07, 0x01]),  # Address 7, P0
        )
        mock_station.send_xbus_command.return_value = response_msg

        turnout = await Turnout.control(mock_station, address=7)

        assert isinstance(turnout, Turnout)
        assert turnout.address == 7

    @pytest.mark.asyncio
    async def test_control_requests_initial_state(self, mock_station):
        """Test that control() fetches initial state to register with station."""
        response_msg = XBusMessage(
            x_header=XBUS_TURNOUT_INFO,
            dbs=bytes([0x00, 0x03, 0x02]),  # Address 3, P1
        )
        mock_station.send_xbus_command.return_value = response_msg

        await Turnout.control(mock_station, address=3)

        mock_station.send_xbus_command.assert_called_once()

    @pytest.mark.asyncio
    async def test_control_suppresses_timeout(self, mock_station):
        """Test that control() succeeds even if the initial get_state times out."""
        mock_station.send_xbus_command.side_effect = TimeoutError

        turnout = await Turnout.control(mock_station, address=10)

        assert turnout.address == 10

    def test_repr(self, mock_station):
        """Test string representation."""
        turnout = Turnout(mock_station, address=42)

        assert "42" in repr(turnout)

    @pytest.mark.asyncio
    async def test_switch_p0_queue_mode(self, mock_station):
        """Test switching to P0 in queue mode."""
        turnout = Turnout(mock_station, address=3)
        await turnout.switch(TurnoutPosition.P0)

        mock_station.send_xbus_command.assert_called_once()
        msg = mock_station.send_xbus_command.call_args[0][0]

        assert msg.x_header == XBUS_SET_TURNOUT
        assert msg.dbs[0] == 0x00  # Address MSB
        assert msg.dbs[1] == 0x03  # Address LSB
        # P0 -> output 0, activate, queue: 1_1_0_1_0_0_0 = 0xA8
        assert msg.dbs[2] == 0xA8

    @pytest.mark.asyncio
    async def test_switch_p1_queue_mode(self, mock_station):
        """Test switching to P1 in queue mode."""
        turnout = Turnout(mock_station, address=3)
        await turnout.switch(TurnoutPosition.P1)

        msg = mock_station.send_xbus_command.call_args[0][0]

        # P1 -> output 1, activate, queue: 1_1_0_1_0_0_1 = 0xA9
        assert msg.dbs[2] == 0xA9

    @pytest.mark.asyncio
    async def test_switch_p0_immediate_mode(self, mock_station):
        """Test switching to P0 in immediate mode sends activate then deactivate."""
        turnout = Turnout(mock_station, address=3)

        with patch("z21aio.turnout.asyncio.sleep", new_callable=AsyncMock) as mock_sleep:
            await turnout.switch(
                TurnoutPosition.P0, queue_mode=False, activate_duration=0.15
            )

        # Should have sent two commands: activate then deactivate
        assert mock_station.send_xbus_command.call_count == 2

        activate_msg = mock_station.send_xbus_command.call_args_list[0][0][0]
        deactivate_msg = mock_station.send_xbus_command.call_args_list[1][0][0]

        # Activate: immediate, output 0, A=1: 1_0_0_1_0_0_0 = 0x88
        assert activate_msg.dbs[2] == 0x88
        # Deactivate: immediate, output 0, A=0: 1_0_0_0_0_0_0 = 0x80
        assert deactivate_msg.dbs[2] == 0x80

        # Sleep should have been called with the duration
        mock_sleep.assert_called_once_with(0.15)

    @pytest.mark.asyncio
    async def test_switch_invalid_position_unknown(self, mock_station):
        """Test that UNKNOWN position raises ValueError."""
        turnout = Turnout(mock_station, address=0)

        with pytest.raises(ValueError, match="P0 or P1"):
            await turnout.switch(TurnoutPosition.UNKNOWN)

    @pytest.mark.asyncio
    async def test_switch_invalid_position_invalid(self, mock_station):
        """Test that INVALID position raises ValueError."""
        turnout = Turnout(mock_station, address=0)

        with pytest.raises(ValueError, match="P0 or P1"):
            await turnout.switch(TurnoutPosition.INVALID)

    @pytest.mark.asyncio
    async def test_activate(self, mock_station):
        """Test low-level activate."""
        turnout = Turnout(mock_station, address=5)
        await turnout.activate(0)

        msg = mock_station.send_xbus_command.call_args[0][0]
        assert msg.x_header == XBUS_SET_TURNOUT
        # Activate, output 0, queue: 0xA8
        assert msg.dbs[2] == 0xA8

    @pytest.mark.asyncio
    async def test_deactivate(self, mock_station):
        """Test low-level deactivate."""
        turnout = Turnout(mock_station, address=5)
        await turnout.deactivate(1)

        msg = mock_station.send_xbus_command.call_args[0][0]
        assert msg.x_header == XBUS_SET_TURNOUT
        # Deactivate, output 1, queue: 1_1_0_0_0_0_1 = 0xA1
        assert msg.dbs[2] == 0xA1

    @pytest.mark.asyncio
    async def test_get_state(self, mock_station):
        """Test getting turnout state."""
        # Create a mock response
        response_msg = XBusMessage(
            x_header=XBUS_TURNOUT_INFO,
            dbs=bytes([0x00, 0x05, 0x01]),  # Address 5, P0
        )
        mock_station.send_xbus_command.return_value = response_msg

        turnout = Turnout(mock_station, address=5)
        state = await turnout.get_state()

        assert state.address == 5
        assert state.position == TurnoutPosition.P0

        # Verify the request was sent with correct response header
        call_args = mock_station.send_xbus_command.call_args
        request_msg = call_args[0][0]
        expected_response_header = call_args[0][1]

        assert request_msg.x_header == XBUS_TURNOUT_INFO
        assert request_msg.dbs == bytes([0x00, 0x05])
        assert expected_response_header == XBUS_TURNOUT_INFO

    @pytest.mark.asyncio
    async def test_get_state_no_response(self, mock_station):
        """Test get_state raises RuntimeError when no response."""
        mock_station.send_xbus_command.return_value = None

        turnout = Turnout(mock_station, address=0)

        with pytest.raises(RuntimeError, match="No response"):
            await turnout.get_state()

    def test_subscribe_state(self, mock_station):
        """Test subscribing to turnout state updates."""
        turnout = Turnout(mock_station, address=5)
        callback = MagicMock()
        turnout.subscribe_state(callback)

        # Verify subscriber was registered
        assert XBUS_TURNOUT_INFO in mock_station._subscribers
        assert len(mock_station._subscribers[XBUS_TURNOUT_INFO]) == 1

        # Simulate receiving a TURNOUT_INFO packet for address 5
        info_msg = XBusMessage(
            x_header=XBUS_TURNOUT_INFO,
            dbs=bytes([0x00, 0x05, 0x02]),  # Address 5, P1
        )
        packet = Packet(header=0x40, data=info_msg.to_bytes())
        mock_station._subscribers[XBUS_TURNOUT_INFO][0](packet)

        callback.assert_called_once()
        state = callback.call_args[0][0]
        assert state.address == 5
        assert state.position == TurnoutPosition.P1

    def test_subscribe_state_filters_address(self, mock_station):
        """Test that subscription filters by turnout address."""
        turnout = Turnout(mock_station, address=5)
        callback = MagicMock()
        turnout.subscribe_state(callback)

        # Simulate receiving a TURNOUT_INFO packet for a different address
        info_msg = XBusMessage(
            x_header=XBUS_TURNOUT_INFO,
            dbs=bytes([0x00, 0x0A, 0x01]),  # Address 10, P0
        )
        packet = Packet(header=0x40, data=info_msg.to_bytes())
        mock_station._subscribers[XBUS_TURNOUT_INFO][0](packet)

        callback.assert_not_called()

    def test_subscribe_state_ignores_short_packets(self, mock_station):
        """Test that GET_TURNOUT_INFO packets (2 DB bytes) are ignored."""
        turnout = Turnout(mock_station, address=5)
        callback = MagicMock()
        turnout.subscribe_state(callback)

        # Simulate receiving a GET_TURNOUT_INFO packet (only 2 DB bytes)
        get_msg = XBusMessage(
            x_header=XBUS_TURNOUT_INFO,
            dbs=bytes([0x00, 0x05]),  # Only 2 bytes = GET request
        )
        packet = Packet(header=0x40, data=get_msg.to_bytes())
        mock_station._subscribers[XBUS_TURNOUT_INFO][0](packet)

        callback.assert_not_called()



"""
Z21 protocol message constants and XBus message handling.

Contains LAN headers, XBus headers, and the XBusMessage class
for handling XBus protocol messages with XOR checksums.
"""

from dataclasses import dataclass
from functools import reduce
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .types import DccThrottleSteps, FunctionAction

# LAN Headers (message types)
LAN_GET_SERIAL_NUMBER = 0x10
LAN_LOGOFF = 0x30
LAN_DISCOVER_DEVICES = 0x35
LAN_XBUS_HEADER = 0x40
LAN_SET_BROADCASTFLAGS = 0x50
LAN_GET_BROADCASTFLAGS = 0x51
LAN_SYSTEMSTATE_DATACHANGED = 0x84
LAN_SYSTEMSTATE_GETDATA = 0x85

# XBus Headers (command types within XBus messages)
XBUS_SET_TRACK_POWER = 0x21
XBUS_GET_VERSION = 0x21
XBUS_GET_VERSION_REPLY = 0x63
XBUS_BC_TRACK_POWER = 0x61
XBUS_BC_TRACK_POWER_OFF_DB0 = 0x00  # DB0 value in power-off broadcast (2.7)
XBUS_BC_TRACK_POWER_ON_DB0 = 0x01   # DB0 value in power-on broadcast (2.8)
XBUS_LOCO_GET_INFO = 0xE3
XBUS_LOCO_DRIVE = 0xE4
XBUS_LOCO_INFO = 0xEF
XBUS_TURNOUT_INFO = 0x43
XBUS_SET_TURNOUT = 0x53
XBUS_GET_FIRMWARE_VERSION = 0xF1
XBUS_GET_FIRMWARE_VERSION_REPLY = 0xF3

# Broadcast flags
BROADCAST_LOCO_INFO = 0x00000001
BROADCAST_TURNOUT_INFO = 0x00000001  # Same as LOCO_INFO
BROADCAST_RAILCOM_SUBSCRIBED = 0x00000004  # RailCom data for subscribed locos
BROADCAST_SYSTEMSTATE = 0x00000100
BROADCAST_RAILCOM_ALL = 0x00040000  # All RailCom data (FW 1.29+)

# RailCom LAN Headers
LAN_RAILCOM_DATACHANGED = 0x88
LAN_RAILCOM_GETDATA = 0x89


@dataclass
class XBusMessage:
    """
    XBus protocol message.

    XBus messages are encapsulated within LAN packets with header 0x40.
    Format: [x_header][data_bytes...][xor_checksum]

    The XOR checksum is calculated as XOR of x_header and all data bytes.

    Attributes:
        x_header: Command type byte
        dbs: Data bytes (variable length)
    """

    x_header: int
    dbs: bytes = b""

    @property
    def xor(self) -> int:
        """Calculate XOR checksum of x_header and all data bytes."""
        return reduce(lambda acc, x: acc ^ x, self.dbs, self.x_header)

    def to_bytes(self) -> bytes:
        """
        Serialize to bytes with XOR checksum.

        Returns:
            Bytes: [x_header][data_bytes][xor]
        """
        return bytes([self.x_header]) + self.dbs + bytes([self.xor])

    @classmethod
    def from_bytes(cls, data: bytes) -> "XBusMessage":
        """
        Parse XBus message from bytes with XOR validation.

        Args:
            data: Raw XBus message bytes

        Returns:
            XBusMessage instance

        Raises:
            ValueError: If XOR checksum is invalid
        """
        if len(data) < 2:
            raise ValueError(f"XBusMessage requires at least 2 bytes, got {len(data)}")

        x_header = data[0]
        dbs = data[1:-1]
        received_xor = data[-1]

        msg = cls(x_header=x_header, dbs=dbs)

        if msg.xor != received_xor:
            raise ValueError(
                f"XBus XOR mismatch: calculated 0x{msg.xor:02X}, "
                f"received 0x{received_xor:02X}"
            )

        return msg

    @classmethod
    def get_firmware_version(cls) -> "XBusMessage":
        """Create command to request firmware version."""
        return cls(x_header=XBUS_GET_FIRMWARE_VERSION, dbs=bytes([0x0A]))

    @classmethod
    def get_version(cls) -> "XBusMessage":
        """Create command to request X-BUS version and command station ID."""
        return cls(x_header=XBUS_GET_VERSION, dbs=bytes([0x21]))

    @classmethod
    def track_power_on(cls) -> "XBusMessage":
        """Create command to turn on track power."""
        return cls(x_header=XBUS_SET_TRACK_POWER, dbs=bytes([0x81]))

    @classmethod
    def track_power_off(cls) -> "XBusMessage":
        """Create command to turn off track power (emergency stop)."""
        return cls(x_header=XBUS_SET_TRACK_POWER, dbs=bytes([0x80]))

    @classmethod
    def loco_get_info(cls, address: int) -> "XBusMessage":
        """
        Create command to request locomotive state.

        Args:
            address: DCC locomotive address (1-9999)

        Returns:
            XBusMessage for getting locomotive info
        """
        addr_msb = (address >> 8) & 0xFF
        addr_lsb = address & 0xFF

        # For addresses >= 128, set the high bits
        if address >= 128:
            addr_msb |= 0xC0

        return cls(x_header=XBUS_LOCO_GET_INFO, dbs=bytes([0xF0, addr_msb, addr_lsb]))

    @classmethod
    def loco_drive(
        cls, address: int, steps: "DccThrottleSteps", speed_byte: int
    ) -> "XBusMessage":
        """
        Create command to drive a locomotive.

        Args:
            address: DCC locomotive address (1-9999)
            steps: Throttle step mode (14/28/128)
            speed_byte: Speed value with direction bit 7
                       (0x00 = stop, 0x01 = emergency stop)

        Returns:
            XBusMessage for driving locomotive
        """
        addr_msb = (address >> 8) & 0xFF
        addr_lsb = address & 0xFF

        # For addresses >= 128, set the high bits
        if address >= 128:
            addr_msb |= 0xC0

        return cls(
            x_header=XBUS_LOCO_DRIVE,
            dbs=bytes([steps.to_speed_byte(), addr_msb, addr_lsb, speed_byte]),
        )

    @classmethod
    def loco_function(
        cls, address: int, function: int, action: "FunctionAction"
    ) -> "XBusMessage":
        """
        Create command to control a locomotive function.

        Args:
            address: DCC locomotive address (1-9999)
            function: Function number (0-31)
            action: Action to perform (OFF, ON, TOGGLE)

        Returns:
            XBusMessage for controlling locomotive function

        Raises:
            ValueError: If function is not 0-31
        """
        if not 0 <= function <= 31:
            raise ValueError(f"Function must be 0-31, got {function}")

        addr_msb = (address >> 8) & 0xFF
        addr_lsb = address & 0xFF

        # For addresses >= 128, set the high bits
        if address >= 128:
            addr_msb |= 0xC0

        # Function byte: TT NNNNNN
        # TT = action type (bits 7-6)
        # NNNNNN = function number (bits 5-0)
        function_byte = (int(action) << 6) | function

        return cls(
            x_header=XBUS_LOCO_DRIVE,
            dbs=bytes([0xF8, addr_msb, addr_lsb, function_byte]),
        )

    @classmethod
    def get_turnout_info(cls, address: int) -> "XBusMessage":
        """Create command to request turnout state.

        Args:
            address: Turnout function address (0-2047)

        Returns:
            XBusMessage for getting turnout info
        """
        addr_msb = (address >> 8) & 0xFF
        addr_lsb = address & 0xFF
        return cls(x_header=XBUS_TURNOUT_INFO, dbs=bytes([addr_msb, addr_lsb]))

    @classmethod
    def set_turnout(
        cls,
        address: int,
        output: int,
        activate: bool,
        queue_mode: bool = True,
    ) -> "XBusMessage":
        """Create command to switch a turnout output.

        Args:
            address: Turnout function address (0-2047)
            output: Output number (0 or 1)
            activate: True to activate, False to deactivate
            queue_mode: True for queue mode (Z21 handles timing),
                       False for immediate (client handles timing)

        Returns:
            XBusMessage for switching turnout

        Raises:
            ValueError: If output is not 0 or 1
        """
        if output not in (0, 1):
            raise ValueError(f"Output must be 0 or 1, got {output}")

        addr_msb = (address >> 8) & 0xFF
        addr_lsb = address & 0xFF

        # DB2 byte format: 1_Q_0_A_0_0_P
        db2 = 0x80  # bit 7 always 1
        if queue_mode:
            db2 |= 0x20  # bit 5 = Q
        if activate:
            db2 |= 0x08  # bit 3 = A
        db2 |= output & 0x01  # bit 0 = P

        return cls(x_header=XBUS_SET_TURNOUT, dbs=bytes([addr_msb, addr_lsb, db2]))

    def __repr__(self) -> str:
        return (
            f"XBusMessage(x_header=0x{self.x_header:02X}, "
            f"dbs={self.dbs.hex()}, xor=0x{self.xor:02X})"
        )
